/*
 GBCartRead
 Version: 1.1
 Author: Alex from Inside Gadgets (http://www.insidegadgets.com)
 Created: 18/03/2011
 Last Modified: 21/03/2011
 
 Read all ROM banks from a Gameboy Cartridge.
 
 */
 
int latchPin = 10;
int dataPin = 11;
int clockPin = 12;
int rdPin = A5;
int wrPin = 13;

void setup() {
  Serial.begin(57600);
  pinMode(latchPin, OUTPUT);
  pinMode(clockPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
  pinMode(rdPin, OUTPUT);
  pinMode(wrPin, OUTPUT);
  pinMode(2, INPUT);
  pinMode(3, INPUT);
  pinMode(4, INPUT);
  pinMode(5, INPUT);
  pinMode(6, INPUT);
  pinMode(7, INPUT);
  pinMode(8, INPUT);
  pinMode(9, INPUT);
}

void loop() {
  
  // Wait for serial input
  while (Serial.available() <= 0) {
    delay(200);
  }
  
  // Read Cartridge Header
  char gameTitle[15];
  for (int addr = 0x0134; addr <= 0x143; addr++) {
    gameTitle[(addr-0x0134)] = (char) readbank0Address(addr);
  }
  gameTitle[14] = '\0';
  int cartridgeType = readbank0Address(0x0147);
  int romSize = readbank0Address(0x0148);
  int ramSize = readbank0Address(0x0149);
  int romBanks = 2;
  if (romSize == 1) { romBanks = 4; } 
  if (romSize == 2) { romBanks = 8; } 
  if (romSize == 3) { romBanks = 16; } 
  if (romSize == 4) { romBanks = 32; } 
  if (romSize == 5 && (cartridgeType == 1 || cartridgeType == 2 || cartridgeType == 3)) { romBanks = 63; } 
  else if (romSize == 5) { romBanks = 64; } 
  if (romSize == 6 && (cartridgeType == 1 || cartridgeType == 2 || cartridgeType == 3)) { romBanks = 125; } 
  else if (romSize == 6) { romBanks = 128; }
  if (romSize == 7) { romBanks = 256; }
  if (romSize == 82) { romBanks = 72; }
  if (romSize == 83) { romBanks = 80; }
  if (romSize == 84) { romBanks = 96; }
  
  // Decode input
  char readInput[10];
  int readCount = 0;
  while (Serial.available() > 0) {
    char c = Serial.read();
    readInput[readCount] = c;
    readCount++;
  }
  readInput[readCount] = '\0';
  
  // Cartridge Header
  if (strstr(readInput, "HEADER")) {
    Serial.println(gameTitle);
    Serial.println(cartridgeType);
    Serial.println(romSize);
    Serial.println(ramSize);
  }
  
  // Dump ROM
  else if (strstr(readInput, "READROM")) {
    Serial.println("START");
    
    unsigned int addr = 0;
    
    // Read x number of banks
    for (int y = 1; y < romBanks; y++) {
      selectROMbank(y);

      if (y > 1) {
        addr = 0x4000;
      }
    
      for (; addr <= 0x7FFF; addr++) {  
        digitalWrite(latchPin, LOW);
        shiftOut(dataPin, clockPin, MSBFIRST, (addr >> 8));
        shiftOut(dataPin, clockPin, MSBFIRST, (addr & 0xFF));
        digitalWrite(latchPin, HIGH);
        delayMicroseconds(50);
        
        byte bval = 0;
        for (int z = 9; z >= 2; z--) {
          if (digitalRead(z) == HIGH) {
            bitWrite(bval, (z-2), HIGH);
          }
        }
        Serial.println(bval, DEC);
      }
    }
    Serial.println("END");
  }
}

// Select the ROM bank by writing the bank number on the D0-D7 pins
void selectROMbank(int bank) {
  digitalWrite(rdPin, HIGH); // RD 1
  digitalWrite(wrPin, LOW); // WR 0
  delay(5);
  
  pinMode(2, OUTPUT);
  pinMode(3, OUTPUT);
  pinMode(4, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);
  pinMode(7, OUTPUT);
  pinMode(8, OUTPUT);
  pinMode(9, OUTPUT);
  
  unsigned int x = 0x2100; // Address where we write to
  digitalWrite(latchPin, LOW);
  shiftOut(dataPin, clockPin, MSBFIRST, (x >> 8));
  shiftOut(dataPin, clockPin, MSBFIRST, (x & 0xFF));
  digitalWrite(latchPin, HIGH);
  delay(1);
  
  // Select the bank
  if (bank & 1) { 
    digitalWrite(2, HIGH);
  }
  if (bank & 2) { 
    digitalWrite(3, HIGH);
  }
  if (bank & 4) { 
    digitalWrite(4, HIGH);
  }
  if (bank & 8) { 
    digitalWrite(5, HIGH);
  }
  if (bank & 16) { 
    digitalWrite(6, HIGH);
  }
  if (bank & 32) { 
    digitalWrite(7, HIGH);
  }
  if (bank & 64) { 
    digitalWrite(8, HIGH);
  }
  if (bank & 128) { 
    digitalWrite(9, HIGH);
  }
  delay(5);
  
  // Set back to reading ROM
  digitalWrite(rdPin, LOW); // RD 0
  digitalWrite(wrPin, HIGH); // WR 1
  
  // Reset outputs to LOW
  for (int l = 2; l <= 9; l++) {
    digitalWrite(l, LOW);
  }
  
  pinMode(2, INPUT);
  pinMode(3, INPUT);
  pinMode(4, INPUT);
  pinMode(5, INPUT);
  pinMode(6, INPUT);
  pinMode(7, INPUT);
  pinMode(8, INPUT);
  pinMode(9, INPUT);
  delay(5);
}

int readbank0Address(int address) {
  digitalWrite(rdPin, LOW); // RD 0
  digitalWrite(wrPin, HIGH); // WR 1
  delay(5);
      
  digitalWrite(latchPin, LOW);
  shiftOut(dataPin, clockPin, MSBFIRST, (address >> 8));
  shiftOut(dataPin, clockPin, MSBFIRST, (address & 0xFF));
  digitalWrite(latchPin, HIGH);
  delay(1);
  
  byte bval = 0;
  for (int z = 9; z >= 2; z--) {
    if (digitalRead(z) == HIGH) {
      bitWrite(bval, (z-2), HIGH);
    }
  }
  return bval;
}
